package com.bonait.bnframework.common.modbus;

import static com.bonait.bnframework.common.modbus.DataConvertLib.ByteToShort;
import static com.bonait.bnframework.common.modbus.DataConvertLib.BytesToFloat;
import static com.bonait.bnframework.common.modbus.DataConvertLib.BytesToInt;
import static com.bonait.bnframework.common.modbus.DataConvertLib.GetAddress;
import static com.bonait.bnframework.common.modbus.DataConvertLib.IntToShorts;

import android.util.Log;

import com.bonait.bnframework.business.ConfigData;
import com.bonait.bnframework.common.constant.ConfigName;
import com.bonait.bnframework.common.helper.DataFormat;
import com.bonait.bnframework.common.helper.I.IReadCallBack;
import com.bonait.bnframework.common.helper.I.IRun;
import com.bonait.bnframework.common.helper.I.IThread;
import com.bonait.bnframework.common.helper.I.IWriteCallBack;
import com.bonait.bnframework.common.helper.MessageLog;
import com.bonait.bnframework.common.helper.ThreadManager;
import com.bonait.bnframework.common.utils.ToastUtils;
import com.licheedev.modbus4android.ModbusCallback;
import com.licheedev.modbus4android.ModbusParam;
import com.licheedev.modbus4android.ModbusRespException;
import com.licheedev.modbus4android.param.SerialParam;
import com.licheedev.modbus4android.param.TcpParam;
import com.serotonin.modbus4j.exception.ModbusInitException;
import com.serotonin.modbus4j.exception.ModbusTransportException;
import com.serotonin.modbus4j.msg.ReadCoilsResponse;
import com.serotonin.modbus4j.msg.ReadHoldingRegistersResponse;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ExecutionException;

import kotlin.UInt;
import kotlin.UShort;

public class ModbusMaster implements IRead,IWrite,IModbusMaster {

    private static volatile ModbusMaster instance = null;
    public static ModbusMaster get() {
        ModbusMaster manager = instance;
        if (manager == null) {
            synchronized (ModbusMaster.class) {
                manager = instance;
                if (manager == null) {
                    manager = new ModbusMaster();
                    instance = manager;
                }
            }
        }
        return manager;
    }
    private ModbusMaster() {
    }

    /**
     * Ping PLC地址是否通畅
     * @param address
     * @param pingTimes  ping的次数
     * @param timeOut  超时时间 10
     * @return
     */
    private  boolean ping2(String address, int pingTimes, int timeOut) {
        Process process = null;
        try {
            process = Runtime.getRuntime().exec( "ping "  + "-c " + pingTimes + " -w " + timeOut+ " "+address);
            InputStreamReader r = new InputStreamReader(process.getInputStream());
            LineNumberReader returnData = new LineNumberReader(r);
            String returnMsg="";
            String line = "";
            while ((line = returnData.readLine()) != null) {
                System.out.println(line);
                returnMsg += line;
            }

            if(returnMsg.indexOf("100% packet loss")!=-1){
                System.out.println("与 " +address +" 连接不畅通.");
                //ToastUtils.info("与 " +address +" 连接不畅通.");
                return  false;
            } else{
                System.out.println("与 " +address +" 连接畅通.");
                //ToastUtils.info("与 " +address +" 连接不畅通.");
                return true;
            }
        } catch (IOException e) {
            // e.printStackTrace();
        }
        return false;

    }

    public boolean IsConnected=false;

    /**
     * TCP 连接，连接时请用多线程的方式，否者有可能无法连接
     * @param ip
     * @param port
     * @param result
     * @throws InterruptedException
     */
    @Override
    public void TcpConnect(String ip, int port, IOperateResult result)throws InterruptedException {
        MessageLog.ShowInfo("开始验证网络："+ip );
        boolean status = false;
        while (!status) {
            try {
                //status为0则代表通，为1则代表不通。
                status =ping2(ip,1,1);
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                MessageLog.ShowInfo("设备 " + ip + " 网络验证失败");
            } catch (Exception e) {
                MessageLog.ShowInfo("设备 " + ip + " 网络验证失败");
            }
        }
        MessageLog.ShowInfo("设备 " + ip + " 网络验证成功");
        ModbusParam param= TcpParam.create(ip, port)
                .setTimeout(1000)
                .setRetries(0)
                .setEncapsulated(false)
                .setKeepAlive(true);
        while (IsConnected==false)
        {
            MessageLog.ShowInfo("Modbus TCP 1111" );
            ModbusBasis.get().init(param, new ModbusCallback<com.serotonin.modbus4j.ModbusMaster>() {
                @Override
                public void onSuccess(com.serotonin.modbus4j.ModbusMaster modbusMaster) {
                    IsConnected=true;
                    if(result!=null) result.onSuccess();
                    MessageLog.ShowInfo("Modbus TCP 连接成功" );
                }
                @Override
                public void onFailure(Throwable tr) {
                    IsConnected = false;
                    if(result!=null) result.onFailure(tr.getMessage());
                    MessageLog.ShowInfo("Modbus TCP 连接失败" );
                }
                @Override
                public void onFinally() {
                    if(result!=null) result.onFinally();
                }
            });
           if(IsConnected==false) Thread.sleep(5000);
        }
    }

    /**
     * RTU 串口连接，建议连接时用多线程的方式，否者有可能无法连接
     * @param com
     * @param baudRate
     * @param result
     * @throws InterruptedException
     */
    @Override
    public void RtuConnect(String com, int baudRate, IOperateResult result)throws InterruptedException {
        MessageLog.ShowInfo("开始连接Modbus RTU" );
        ModbusParam param=SerialParam.create(com, baudRate) // 串口地址和波特率
                .setDataBits(8) // 数据位
                .setParity(0) // 校验位
                .setStopBits(1) // 停止位
                .setTimeout(500).setRetries(0); // 不重试
        while (IsConnected==false)
        {
            ModbusBasis.get().init(param, new ModbusCallback<com.serotonin.modbus4j.ModbusMaster>() {
                @Override
                public void onSuccess(com.serotonin.modbus4j.ModbusMaster modbusMaster) {
                    IsConnected=true;
                    if(result!=null) result.onSuccess();
                    MessageLog.ShowInfo("Modbus RTU 连接成功" );
                }
                @Override
                public void onFailure(Throwable tr) {
                    IsConnected = false;
                    if(result!=null) result.onFailure(tr.getMessage());
                    MessageLog.ShowInfo("Modbus RTU 连接失败" );
                }
                @Override
                public void onFinally() {
                    if(result!=null) result.onFinally();
                }
            });
            if(IsConnected==false) Thread.sleep(5000);
        }
    }

    @Override
    public void ReadBool(String address, int length, int slaveId, IReadCallBack<boolean[]> callback) {
        if (!IsConnected) {
            return;
        }
        new ExceptionServer(address, length, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ReadCoilsResponse res = ModbusBasis.get().syncReadCoil(slaveId, add, length);
                boolean[] data = res.getBooleanData();
                boolean[] result = Arrays.copyOfRange(data, 0, length);
                if (callback != null) callback.onSuccess(result);
            }

            @Override
            public void onFailure(String error){
            }
        });
    }

    @Override
    public void ReadShort(String address, int length,int slaveId, IReadCallBack<short[]> callback) {
        if (!IsConnected) {
            return;
        }
        new ExceptionServer(address, length, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ReadHoldingRegistersResponse res = ModbusBasis.get().syncReadHoldingRegisters(slaveId, add, length);
                short[] data = res.getShortData();
                if (data.length == length) {
                    if (callback != null) callback.onSuccess(data);
                }
            }

            @Override
            public void onFailure(String error){

            }
        });
    }

    @Override
    public void ReadInt(String address, int length, int slaveId, DataFormat df, IReadCallBack<int[]> callback) {
        if (!IsConnected) {
            return;
        }
        new ExceptionServer(address, length, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ReadHoldingRegistersResponse res = ModbusBasis.get().syncReadHoldingRegisters(slaveId, add, length * 2);
                byte[] data = res.getData();
                int[] tempValues = new int[length];
                if(data.length==4*length)
                {
                    for (int i = 0; i < length; i++) {
                        byte[] tempData = new byte[4];
                        for (int m = 0; m < 4; m++) {
                            tempData[m] = data[i * 4 + m];
                        }
                        tempValues[i] = BytesToInt(tempData, df);
                    }
                   if (callback != null) callback.onSuccess(tempValues);
                }else{
                    MessageLog.ShowError("ReadInt onFailure，数据长度不匹配,读取长度："+length+"返回长度："+data.length );
                }
            }

            @Override
            public void onFailure(String error){

            }
        });
    }

    @Override
    public void ReadInt(String address, int length, int slaveId, IReadCallBack<int[]> callback) {
        if (!IsConnected) {
            return;
        }
        ReadInt(address, length, slaveId,DataFormat.BADC, callback);
    }

    @Override
    public void ReadByte(String address, int length, int slaveId, IReadCallBack<byte[]> callback) {
        if (!IsConnected) {
            return;
        }
        new ExceptionServer(address, length, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
               int tempLength=length;
                if (length%2!=0){
                    tempLength++;
               }
                ReadHoldingRegistersResponse res = ModbusBasis.get().syncReadHoldingRegisters(slaveId, add, tempLength );
                byte[] data = res.getData();
                if(data.length==tempLength)
                {
                    byte[] result = Arrays.copyOfRange(data, 0, length);
                    if (callback != null) callback.onSuccess(result);
                }else
                {
                    MessageLog.ShowError("ReadInt onFailure，数据长度不匹配,读取长度："+length+"返回长度："+data.length );
                }
            }

            @Override
            public void onFailure(String error){

            }
        });
    }

    @Override
    public void ReadString(String address, int length,  int slaveId, IReadCallBack<String> callback) {
        if (!IsConnected) {
            return;
        }
        new ExceptionServer(address, length, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ReadHoldingRegistersResponse res = ModbusBasis.get().syncReadHoldingRegisters(slaveId, add, length);
                byte[] data = res.getData();
                try {
                    if (data.length ==length*2) {
                        if (callback != null) callback.onSuccess(new String(data, "UTF-8").trim());
                    }
                } catch (UnsupportedEncodingException ex) {
                }
            }

            @Override
            public void onFailure(String error){

            }
        });
    }

    @Override
    public void ReadFloat(String address, int length, int slaveId, DataFormat df, IReadCallBack<float[]> callback) {
        if (!IsConnected) {
            return;
        }
        new ExceptionServer(address, length, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ReadHoldingRegistersResponse res = ModbusBasis.get().syncReadHoldingRegisters(slaveId, add, length*2);
                byte[] data = res.getData();
                float[] tempValues = new float[length];
                for (int i = 0; i < length; i++) {
                    byte[] tempData = new byte[4];
                    for (int m = 0; m < 4; m++) {
                        tempData[m] = data[i * 4 + m];
                    }
                    tempValues[i] = BytesToFloat(tempData, df);
                }
                if (tempValues.length == length) {
                    if (callback != null) callback.onSuccess(tempValues);
                }
            }

            @Override
            public void onFailure(String error){

            }
        });
    }

    @Override
    public void ReadFloat(String address, int length, int slaveId, IReadCallBack<float[]> callback) {
        if (!IsConnected) {
            return;
        }
        ReadFloat(address,length,slaveId,DataFormat.BADC,callback);
    }




    @Override
    public void WriteBool(String address, boolean value, int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ModbusBasis.get().syncWriteCoil(slaveId, add, value);
                if (callback != null)callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteBool(String address, boolean[] value,  int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ModbusBasis.get().syncWriteCoils(slaveId, add, value);
                if (callback != null)callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteShort(String address, short value,   int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                short[] send = new short[1];
                send[0] = value;
                ModbusBasis.get().syncWriteRegisters(slaveId, add, send);
                if (callback != null)callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteShort(String address, short[] value,   int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                ModbusBasis.get().syncWriteRegisters(slaveId, add, value);
                if (callback != null)callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteInt(String address, int value,  int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                short[] send = IntToShorts(value);
                ModbusBasis.get().syncWriteRegisters(slaveId, add, send);
                if (callback != null) callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteInt(String address, int[] value,  int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
              short[] send=new short[value.length*2];
               for(int i=0;i<value.length;i=i+2)
               {
                   short[] temp = IntToShorts(value[i]);
                  if (temp.length==2)
                  {
                      send[i]=temp[0];
                      send[i+1]=temp[1];
                  }
               }
                ModbusBasis.get().syncWriteRegisters(slaveId, add,send);
                if (callback != null) callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteString(String address, String value,   int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
               byte[] bytes= value.getBytes();
                short[] send =ByteToShort(bytes);
                ModbusBasis.get().syncWriteRegisters(slaveId, add, send);
                if (callback != null) callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteFloat(String address, float value, int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                int intBits = Float.floatToRawIntBits(value);
                short[] send = IntToShorts(intBits);
                ModbusBasis.get().syncWriteRegisters(slaveId, add, send);
                if (callback != null) callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }

    @Override
    public void WriteFloat(String address, float[] value,  int slaveId, IWriteCallBack callback) {
        if (!IsConnected) {
            if (callback != null) callback.onFailure("设备未连接");
            return;
        }
        new ExceptionServer(address, 1, new IExceptionHandling() {
            @Override
            public void onSuccess(int add) throws InterruptedException, ExecutionException, ModbusTransportException, ModbusInitException, ModbusRespException {
                short[] send=new short[value.length*2];
                for(int i=0;i<value.length;i=i+2)
                {
                    int intBits = Float.floatToRawIntBits(value[i]);
                    short[] temp = IntToShorts(intBits);
                    if (temp.length==2)
                    {
                        send[i]=temp[0];
                        send[i+1]=temp[1];
                    }
                }
                ModbusBasis.get().syncWriteRegisters(slaveId, add, send);
                if (callback != null) callback.onSuccess();
            }

            @Override
            public void onFailure(String error){
                if (callback != null)callback.onFailure(error);
            }
        });
    }




}
